
-- Simple Customizable Database derived from Euphoria 1.4 example MyData.ex.
--Modified by Art Adamson, euclid@horandata.net, June 18, 1996 
--and again 8/2001 to fix a bug which caused crashes for repeated saves/mods.
--  User selected data file name (.db is the extension) so many
--    different sets of data can be used.
--  One-line data presentation/limitation, 79 character line
--  Specifying the # of fields, 6 maximum
--  Sorting on any field
--  Optional printing the "found" items to a printer 
--    or saving (overwrite or append) to a file named selected.txt.         .
--  Field width 1 to 12 maximum, but the max 12 can be automatically exceeded
--    for the last field if less than 6 fields are used. The last field is
--    limited (truncated) to enough chars to fill the remaining spaces in the
--    line up to a total of 79 spaces. The other fields are truncated at 12
--    characters.
--  
--  User input is required/permitted for the data field names such as :
--    Last Name, First Name, Phone Number, etc.
--Use command line  ex 6x12db.ex. You will be prompted for a file name (no ext)
--  If the file name exists, it will be used; else you will be asked for the
--    input needed to create a new file.
--Use ctl c to abort. A backup db file is maintained and updated when the db
--file is changed and saved.
--An abort will leave a 2 byte backup file. Delete it or it will cause an 
--    error if the same name is called again.
--All of above can be further customized from the following source code.
--The prog requires the Euphoria include files get.e and graphics.e    .
--The prog requires the include file sortcols.e which was derived from sort.e
--  by A. P. Adamson to provide sorting on any item of a sequence rather than
--  only the first.
    
include get.e
include sortcols.e      --Adamson mod of sort.e to sort on any column
include graphics.e
    
    sequence fields, DB_NAME, BACK_NAME, SELECTED_NAME
    integer tabpos                      --Used to align columns
    tabpos = 1
    fields = {}             --Contains each field name for 1 data set
    DB_NAME   = {}          --DB storage file name
    BACK_NAME = {}          --backup DB storage file after a change
    SELECTED_NAME = {}      --Saves selected items from DB to a file
    integer numfields, max_line_length, items_per_screen, field_width
    object lin
    
constant KEYBOARD = 0, SCREEN   = 1, ERROR    = 2, TRUE = 1
    max_line_length = 79
    items_per_screen  = 20      --lines (items) in each list screen
    field_width     = 12        --useable width, except last field is wider

function upper(sequence name)  -- convert to upper case
    for i = 1 to length(name) do
	if name[i] >= 'a' and name[i] <= 'z' then
	    name[i] = name[i] + 'A' - 'a'
	end if
    end for
    return name
end function    --upper

procedure getdbname()
    integer index
    index = 0
    puts(SCREEN,
	"What name for DB storage file (I will add the extension .db) ")
    lin = gets(KEYBOARD)
    if find('.', lin)       then  
	index = find('.', lin)
	if index = length(lin) - 1 then
--protect against "filename." as input (added .)
	    lin = lin[1..length(lin) - 2] & lin[length(lin)]
	else
--protect against "filename.db" as input (added .db)
	    lin = lin[1..length(lin) - 4] & lin[length(lin)]
	end if
    end if
    
    DB_NAME = append(DB_NAME,lin[1..length(lin) - 1])
    DB_NAME = append(DB_NAME,".db")
    DB_NAME = DB_NAME[1] & DB_NAME[2]
    BACK_NAME = append(BACK_NAME,lin[1..length(lin) - 1])
    BACK_NAME = append(BACK_NAME,".bak")
    BACK_NAME = BACK_NAME[1] & BACK_NAME[2]
    SELECTED_NAME = append(SELECTED_NAME, "selected.txt")
end procedure       --getdbname

procedure getfields(integer isdb)
    integer len
    if isdb = -1  then
	bk_color(MAGENTA)   text_color(BLACK)
	puts(SCREEN,
    "\nYou may use 1 to 6, fields having 12 characters + a space but the\n")
	puts(SCREEN,
    "LAST field, if fewer than 6 fields are used, can be wider depending\n")
	puts(SCREEN,"on the space left for listing (79 char max line length).")
	text_color(BRIGHT_CYAN)
	puts(SCREEN,"  Use CTL C to abort\n")
	text_color(BLACK)   bk_color(YELLOW)
	puts(SCREEN,"\nThe db can be sorted vs any" & 
	    "field but find & delete work only vs the first.")
	bk_color(BLACK)     text_color(WHITE)
	puts(SCREEN,
	    "\nHow many fields do you want, input 1, 2, 3, 4, 5, or 6?")
	lin = get(KEYBOARD)
	numfields = lin[2] 
	printf(1, "%d", numfields)
	puts(SCREEN,"\n  Type their names (12 char or less)," & 
	    "in order, separated by enter.\n")
	for i = 1 to numfields do
	    lin = gets(KEYBOARD)
	    len = length(lin)    
	    fields = append(fields,lin[1..len-1])
	end for
    end if
end procedure       --getfields
------------------------------------------------------------------------------ 
type file_number(integer x)     return x >= -1                  end type
type record(sequence s)         return length(s) = numfields    end type
    file_number db          -- handle of file containing database
    file_number selected_handle
    sequence database  --, selected       -- the in-memory database
type record_number(integer x)  return x >= 0 and x <= length(database) end type

procedure error(sequence msg)  -- fatal error
    puts(ERROR, '\n' & msg & '\n')  abort(1)  
end procedure       --error

procedure show(file_number f, record rec)
    sequence linecol, line_to_print
    linecol = {}                --holds cursor position info
    line_to_print = {}
    tabpos = 1
    puts(f,"\n")
    if f = SCREEN then          --show for screen
      for i = 1 to numfields do
	linecol = get_position()
	position(linecol[1] - 1 , tabpos )
	puts(f, rec[i])
	tabpos = tabpos + field_width + 1
      end for
    else
--I had trouble making "position" work for printer so avoided it below
--I build a complete print line, then print it
      for i = 1 to numfields do             --show for printer
	line_to_print = line_to_print & rec[i][1..length(rec[i])-1]
	if i != numfields then
	    line_to_print = line_to_print & 
	    repeat(' ', field_width - length(rec[i]) + 2)
	end if
      end for
      puts(f,line_to_print)     
    end if
end procedure   --show

function lookup(sequence name)  -- return a record matching name
    sequence matches
    matches = {}
    name = upper(name)
    for i = 1 to length(database) do
	if compare(name, upper(database[i][1])) = 0 then
	    matches = matches & i
	end if
    end for
    return matches
end function    --lookup

procedure db_add()              -- add a new record to the database
    record rec
    sequence matches, aline
    integer spaces_left
    numfields = length(fields)
    rec = repeat(0, numfields)
    spaces_left = 0
    puts(SCREEN, "\n" & fields[1] & ": ")
    --Special treatment for first field because of look up function, ie
    --  a provision to signal duplicates
    aline = gets(KEYBOARD)
    if length(aline) > field_width then
	rec[1] = aline[1..field_width] & aline[length(aline)]
	puts(SCREEN,"Truncating to 12 characters!\n")
    else 
	rec[1] = aline
    end if
    matches = lookup(rec[1])
    for i = 1 to length(matches) do
	show(SCREEN, database[matches[i]])
    end for
    for i = 2 to numfields - 1 do
	puts(SCREEN, "\n" & fields[i] & ": ")
	aline = gets(KEYBOARD)
	if length(aline) > field_width then
	    rec[i] = aline[1..field_width] & aline[length(aline)]
	    puts(SCREEN,"Truncating to 12 characters!\n")
	else
	    rec[i] = aline  
	end if
    end for
--  Special width allowed for last field to use all available space
	spaces_left = max_line_length - (numfields-1) * (field_width + 1)
	puts(SCREEN, "\nSpaces left are: ")   
	? spaces_left 
	puts(SCREEN,"\n" & fields[numfields] & ": ")
	aline = gets(KEYBOARD)
	if length(aline) > spaces_left then
	    rec[numfields] = aline[1..spaces_left] & aline[length(aline)]
	    puts(SCREEN,"\nCharacters used are: ")
	    ? spaces_left 
	else
	    rec[numfields] = aline  
	end if
    database = append(database, rec)
end procedure       --db_add

procedure db_delete()       -- delete a record, given first field 
    sequence name, answer
    record_number rec_num
    sequence matches
    integer i
    puts(SCREEN, "\n\t" & fields[1] & ": ")
    name = gets(KEYBOARD)
    matches = lookup(name)
    if length(matches) = 0 then
	puts(SCREEN, "\n\tnot found\n")
	return
    end if 
    i = 1
    while i <= length(matches) do
	puts(1,"\n")
	show(SCREEN, database[matches[i]])
	puts(SCREEN, "\nDelete it, hit y or n?\n ")
	answer = gets(KEYBOARD)
	if find('y', answer) then
	    rec_num = matches[i]
	    database = database[1..rec_num-1] & 
		database[rec_num+1..length(database)]
	    exit
	end if
	i = i + 1
    end while
end procedure       --db_delete

procedure db_find()    -- find all records that match value of first field
    file_number printer
    sequence name, matches, answer
    answer = {}
    puts(SCREEN, "\n\t" & fields[1] & ": ")
    name = gets(KEYBOARD)
    matches = lookup(name)
    if length(matches) = 0 then puts(SCREEN, "\n\tnot found\n") 
    end if 
    for i = 1 to length(matches) do show(SCREEN, database[matches[i]])
    end for     
    --option to save-to-file found-items
    puts(SCREEN, 
	"\nDo you want to save to file selected.txt? y or n, default is n ")
    answer = gets(0)
    if answer[1] = 'y' or answer[1] = 'Y'  then
	puts(SCREEN, 
	"\n   Overwrite or append selected.txt? o or a, default is a ")
	answer = {}
	answer = gets(0)
	if answer[1] = 'o' or answer[1] = 'O' then
	    selected_handle = open(SELECTED_NAME[1], "w")
	else
	    selected_handle = open(SELECTED_NAME[1], "a")
	end if
	if selected_handle = -1 then error("Can't save database")
	else
	    puts(selected_handle,"\n\nYour selected items are:")
	    for i = 1 to length(matches) do
		show(selected_handle, database[matches[i]])
	    end for
	    close(selected_handle)
	end if
    end if
    --option to print found items
	puts(SCREEN, "\n\tDo you want to print it? y or n, default is n ")
	answer = {}
	answer = gets(0)
	if answer[1] = 'y' or answer[1] = 'Y' then  
	    printer = open("PRN", "w")
	    if printer = -1 then 
		puts(SCREEN, "Can't open printer device\n")
	    else           
		puts(printer,"\n\nYour selected items are:\n")
		for i = 1 to length(matches) do 
		    show(printer, database[matches[i]])
		end for
		close(printer)
	    end if
	end if
end procedure       --db_find

procedure db_list(file_number f)    -- list the entire database to a device
    sequence sorted_database
    sequence sfieldnum, linecol, dummy
    integer tabpos, i1
    puts(f, "\n\n                   THE DATA BASE CONTAINS:\n")
    printf(1, "The DB has %d ", (numfields))
    puts(SCREEN,"fields. Sort on which one? Input 1, 2, 3, ... (1 is default)")
    sfieldnum = gets(KEYBOARD)
    if length(sfieldnum) = 1 then
	sfieldnum = {49,10}     --default = 1 for first field
    else 
	sfieldnum = sfieldnum[1] & {10}
    end if
    clear_screen()
    bk_color(GREEN)     text_color(BLACK)
    puts(f,"\nDB    " & DB_NAME & "     is ")
    printf(f,"sorted on field # %d. \n", (sfieldnum-48))
    bk_color(BLACK)     text_color(WHITE)
    sorted_database = sortcols(database, (sfieldnum[1] - 48))
    tabpos = 1
    bk_color(RED)       text_color(BLACK)
    if f = SCREEN then
	for i = 1 to numfields do       --show field names
	    puts(f, fields[i] )
	    linecol = get_position()
	    tabpos = tabpos + field_width + 1
	    position(linecol[1], tabpos )
	end for
    else
	for i = 1 to numfields do   --special show field names for printer. I
				--didn't make position work for printer 
	    puts(f, fields[i] )
	    dummy = {}
	    dummy = repeat(' ', field_width - length(fields[i])+1)
	    puts(f,dummy)   --adds spaces as needed to msaintain tab columns
	end for
    end if
    bk_color(BLACK)     text_color(WHITE)
    puts(f,
	"\n--------------------------Data Sets Are:" &
	"----------------------------------------")
    for i = 1 to length(sorted_database) do --show one screen-full at a time
	show(f, sorted_database[i]) 
	if f = SCREEN then
	    i1 = remainder(i,items_per_screen) 
	    if i1 = 0 and i < length(sorted_database)  then 
		bk_color(GREEN)     text_color(BLACK)
		puts(SCREEN, "Hit enter for more")
		bk_color(BLACK)     text_color(WHITE)
		dummy = gets(KEYBOARD) puts(SCREEN,dummy)    
		tabpos = 0
		bk_color(RED)       text_color(BLACK)
		for j = 1 to numfields do
		    puts(f, fields[j] ) 
		    linecol = get_position()
		    tabpos = tabpos + field_width + 1
		    position(linecol[1], tabpos )
		end for
		bk_color(BLACK)     text_color(WHITE)
		puts(SCREEN,"\n")
	    end if
	end if
    end for
    if f = SCREEN then
	bk_color(RED)       text_color(BLACK)
    puts(f,
	"\n--------------------------That Is All!!!" &
	"----------------------------------------")
	
	bk_color(BLACK)     text_color(WHITE)
    end if
	puts(SCREEN,"\n")
end procedure       --db_list

procedure db_create()         -- create a new database
    db = open(DB_NAME, "w")
    database = {}
    print(db, database)
    close(db)
    db = open(DB_NAME, "r")
    if db = -1 then
	error("Couldn't open database")
    end if    
end procedure       --db_create

procedure db_save()         -- save in-memory database to disk file
    system("copy " & DB_NAME & " " & BACK_NAME & " > NUL", 2)
    db = open(DB_NAME, "w")
    if db = -1 then
	system("copy " & BACK_NAME & " " & DB_NAME & " > NUL", 2)
	error("Can't save database")
    end if
    -- we could save space in the file by using puts() to output strings
    -- like "ABC". print() outputs numbers like {65, 66, 67}
    print(db, database)
    puts(db,",")
    print(db, fields)    --Adamson added so structure could be saved
    close(db)
end procedure       --db_save

procedure db_main()
    sequence command
    file_number printer
    getdbname()
    puts(SCREEN,"\nThe dbname is:  " & DB_NAME & "\n")
    db = open(DB_NAME, "r")  
    if db = -1 then
	getfields(-1)
	db_create()
    else
	database = get(db)
	if database[1] != GET_SUCCESS then
	    error(
	    "Couldn't read database. Format must be different than in file")
	end if
	fields = get(db)

	fields = fields[2]
	? fields[1]
	? fields[2]
	database = database[2]
	numfields = length(fields)
    end if
    close(db)
    clear_screen()
    puts(SCREEN, "These operations are useable:\n")
    while TRUE do
	text_color(YELLOW)
	puts(SCREEN, 
	"\n(a)dd, (d)elete, (f)ind, (l)ist, (p)rint, (s)ave, (q)uit: ")
	text_color(WHITE)
	command = gets(KEYBOARD)
	if find('a', command) then
	    db_add()
	elsif find('d', command) then
	    db_delete()
	elsif find('f', command) then
	    db_find()
	elsif find('q', command) then
	    puts(SCREEN, "Save before exit, hit y or n, Default = no?")
	    command = gets(KEYBOARD)
	    if find('y', command) then db_save() exit
	    elsif find('n', command) and length(database) = 0 then 
		system("del " & DB_NAME, 2)  exit 
	    else exit
	    end if
	elsif find('s', command) then
	    db_save()  exit
	elsif find('l', command) then
	    db_list(SCREEN)
	elsif find('p', command) then
	    printer = open("PRN", "w")
	    if printer = -1 then
		puts(SCREEN, "Can't open printer device\n")
	    else
		db_list(printer)
-----------------------------               puts(printer, FORM_FEED)
		close(printer)
	    end if
	else
	    puts(SCREEN, "\nYou forgot to select an operation\n")              
	end if 
    end while
end procedure

db_main()  --db_main    --The real call to action



